//////////////////////////////////////////////////////////////////////////////

template<class PLUGIN_TABLE>
class AviUtlPluginObject : public RefCount<AviUtlPluginObject> {
	LibraryObjectRef _library;
	const char *_name;
	PLUGIN_TABLE *_table;

	enum State { INIT, PROC, EXIT };
	State _state;

protected:
	AviUtlPluginObject(const char *path, const char *name, bool copy, const char *api)
		: _name(name), _table(NULL), _state(INIT)
	{
		if(copy)
			_library = new CopiedLibrary(path);
		else
			_library = new Library(path);

		typedef PLUGIN_TABLE *(__stdcall *GET_PLUGIN_TABLE)();
		if(!_library) return;

		void *func = _library->GetAddress(api);
		if(func == NULL) return;

		_table = static_cast<GET_PLUGIN_TABLE>(func)();
	}

	virtual bool OnInit() = 0;
	virtual bool OnExit() = 0;

public:
	typedef PLUGIN_TABLE PLUGIN_TABLE;

	virtual ~AviUtlPluginObject()
	{}

	bool operator!() const
	{ return _table == NULL; }

	bool Init() {
		if(_table == NULL)
			return false;

		if(_state == INIT && OnInit())
			_state = PROC;

		return _state == PROC;
	}

	bool Exit() {
		if(_table == NULL)
			return false;

		if(_state == PROC && OnExit())
			_state = EXIT;

		return _state == EXIT;
	}

	const LibraryObjectRef& GetLibrary() const
	{ return _library; }

	const char *GetFuncName() const
	{ return _name; }

	const char *GetInformation() const
	{ return (_table->information != NULL) ? _table->information : ""; }

	PLUGIN_TABLE *GetPluginTable() const
	{ return _table; }
};

//////////////////////////////////////////////////////////////////////////////

class ConvertYUY2ToAviUtlYC : public GenericVideoFilter {
	static int ConvertMMX2(PIXEL_YC *yc, const BYTE *yuv, int width) {
		static const __int64 _0080000000800000 = 0x0080000000800000;

		if(width /= 4) __asm {
			pxor		mm7,mm7
			movq		mm6,_0080000000800000
			xor			esi,esi
			mov			ecx,width
			mov			eax,yuv
			mov			ebx,yc
		loopX:
			movq		mm0,[eax+esi*8]
			inc			esi
			pshufw		mm1,mm0,(0<<6)+(0<<4)+(3<<2)+(2<<0)
			punpcklbw	mm0,mm7
			punpcklbw	mm1,mm7
			psubw		mm0,mm6
			psubw		mm1,mm6
			psllw		mm0,4
			psllw		mm1,4
			pshufw		mm2,mm0,(2<<6)+(3<<4)+(1<<2)+(0<<0)
			pshufw		mm3,mm0,(0<<6)+(0<<4)+(3<<2)+(1<<0)
			pshufw		mm4,mm1,(0<<6)+(0<<4)+(1<<2)+(0<<0)
			punpckldq	mm3,mm4
			pshufw		mm5,mm1,(3<<6)+(1<<4)+(2<<2)+(3<<0)
			movntq		[ebx],mm2
			movntq		[ebx+8],mm3
			movntq		[ebx+16],mm5
			add			ebx,24
			loop		loopX
		}

		return width * 4;
	}

	ConvertYUY2ToAviUtlYC(const PClip& clip, IScriptEnvironment *env)
		: GenericVideoFilter(clip)
	{
		vi.width *= 3;
		child->SetCacheHints(CACHE_NOTHING, 0);
	}

	static AVSValue __cdecl Create(AVSValue args, void *user_data, IScriptEnvironment *env) {
		PClip child(args[0].AsClip());
		const VideoInfo& vi = child->GetVideoInfo();

		if(!vi.IsYUY2())
			env->ThrowError("%s: YUY2p", GetName());

		return new ConvertYUY2ToAviUtlYC(child, env);
	}

public:
	static const char *GetName()
	{ return "ConvertYUY2ToAviUtlYC"; }

	static void PluginInit(IScriptEnvironment *env)
	{ env->AddFunction(GetName(), "c", Create, 0); }

	static bool Convert(PIXEL_YC *yc, int ycPitch, const BYTE *yuv, int yuvPitch, int width, int height, long cpu) {
		bool emms = false;

		if(width * sizeof(PIXEL_YC) == ycPitch && width * 2 == yuvPitch) {
			emms = Convert(yc, yuv, width * height, cpu);
		}
		else {
			for(int h = height; h; --h) {
				emms = Convert(yc, yuv, width, cpu);
				reinterpret_cast<BYTE *&>(yc) += ycPitch;
				yuv += yuvPitch;
			}
		}

		return emms;
	}

	static bool Convert(PIXEL_YC *yc, const BYTE *yuv, int width, long cpu) {
		bool emms = false;

		if(cpu & CPUF_INTEGER_SSE) {
			int w = ConvertMMX2(yc, yuv, width);
			width -= w;
			yuv += w * 2;
			yc += w;
			emms = true;
		}

		for(int w = width / 2; w; --w) {
			yc[0].y = yuv[0] << 4;
			yc[1].y = yuv[2] << 4;
			yc[0].cb = yc[1].cb = (yuv[1] - 128) << 4;
			yc[0].cr = yc[1].cr = (yuv[3] - 128) << 4;
			yuv += 4;
			yc += 2;
		}

		return emms;
	}

	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment *env) {
		PVideoFrame srcFrame(child->GetFrame(n, env));
		PVideoFrame dstFrame(env->NewVideoFrame(vi, sizeof(PIXEL_YC)));

		bool flag = Convert(reinterpret_cast<PIXEL_YC *>(dstFrame->GetWritePtr()), dstFrame->GetPitch(),
			srcFrame->GetReadPtr(), srcFrame->GetPitch(), vi.width / 3, vi.height, env->GetCPUFlags());

		if(flag)
			__asm emms

		return dstFrame;
	}
};

//////////////////////////////////////////////////////////////////////////////

class ConvertAviUtlYCToYUY2 : public GenericVideoFilter {
	static int ConvertMMX2(BYTE *yuv, const PIXEL_YC *yc, int width) {
		static const __int64 _0800000008000000 = 0x0800000008000000;

		if(width /= 4) __asm {
			mov			ecx,width
			mov			esi,yc
			shl			ecx,3
			mov			edi,yuv
			lea			ebx,[ecx+ecx*2]
			add			edi,ecx
			add			esi,ebx
			neg			ecx
			neg			ebx
			movq		mm6,_0800000008000000
			pxor		mm7,mm7
		loopX:
			movq		mm0,[esi+ebx]
			movq		mm1,[esi+ebx+8]
			movq		mm2,[esi+ebx+16]
			pshufw		mm1,mm1,(0<<6)+(0<<4)+(3<<2)+(2<<0)
			pshufw		mm2,mm2,(0<<6)+(0<<4)+(0<<2)+(1<<0)
			pshufw		mm0,mm0,(2<<6)+(3<<4)+(1<<2)+(0<<0)
			punpckldq	mm1,mm2
			paddsw		mm0,mm6
			paddsw		mm1,mm6
			psraw		mm0,4
			psraw		mm1,4
			packuswb	mm0,mm7
			packuswb	mm1,mm7
			add			ebx,24
			punpckldq	mm0,mm1
			add			ecx,8
			movntq		[edi+ecx-8],mm0
			jnz			loopX
		}

		return width * 4;
	}

	static int Clamp(int n, int l, int h)
	{ return (n < l) ? l : (n > h) ? h : n; }

	ConvertAviUtlYCToYUY2(const PClip& clip, IScriptEnvironment *env)
		: GenericVideoFilter(clip)
	{
		vi.width /= 3;
		child->SetCacheHints(CACHE_NOTHING, 0);
	}

	static AVSValue __cdecl Create(AVSValue args, void *user_data, IScriptEnvironment *env) {
		PClip child(args[0].AsClip());
		const VideoInfo& vi = child->GetVideoInfo();

		if(!vi.IsYUY2())
			env->ThrowError("%s: YUY2p", GetName());

		return new ConvertAviUtlYCToYUY2(child, env);
	}

public:
	static const char *GetName()
	{ return "ConvertAviUtlYCToYUY2"; }

	static void PluginInit(IScriptEnvironment *env)
	{ env->AddFunction(GetName(), "c", Create, 0); }

	static bool Convert(BYTE *yuv, int yuvPitch, const PIXEL_YC *yc, int ycPitch, int width, int height, long cpu) {
		bool emms = false;

		if(width * 2 == yuvPitch && width * sizeof(PIXEL_YC) == ycPitch) {
			emms = Convert(yuv, yc, width * height, cpu);
		}
		else {
			for(int h = height; h; --h) {
				emms = Convert(yuv, yc, width, cpu);
				yuv += yuvPitch;
				reinterpret_cast<const BYTE *&>(yc) += ycPitch;
			}
		}

		return emms;
	}

	static bool Convert(BYTE *yuv, const PIXEL_YC *yc, int width, long cpu) {
		bool emms = false;

		if(cpu & CPUF_INTEGER_SSE) {
			int w = ConvertMMX2(yuv, yc, width);
			width -= w;
			yc += w;
			yuv += w * 2;
			emms = true;
		}

		for(int w = width / 2; w; --w) {
			yuv[0] = Clamp(yc[0].y,         0, 4095) >> 4;
			yuv[2] = Clamp(yc[1].y,         0, 4095) >> 4;
			yuv[1] = Clamp(yc[0].cb + 2048, 0, 4095) >> 4;
			yuv[3] = Clamp(yc[0].cr + 2048, 0, 4095) >> 4;
			yc += 2;
			yuv += 4;
		}

		return emms;
	}

	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment *env) {
		PVideoFrame srcFrame(child->GetFrame(n, env));
		PVideoFrame dstFrame(env->NewVideoFrame(vi));

		bool flag = Convert(dstFrame->GetWritePtr(), dstFrame->GetPitch(),
			reinterpret_cast<const PIXEL_YC *>(srcFrame->GetReadPtr()),
			srcFrame->GetPitch(), vi.width, vi.height, env->GetCPUFlags());

		if(flag)
			__asm emms

		return dstFrame;
	}
};

//////////////////////////////////////////////////////////////////////////////
